=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2006-08-17 07:52:09 +0000
+++ bzrlib/branch.py	2006-08-21 21:45:16 +0000
@@ -363,7 +363,7 @@
             return None
         if history is None:
             history = self.revision_history()
-        elif revno <= 0 or revno > len(history):
+        if revno <= 0 or revno > len(history):
             raise bzrlib.errors.NoSuchRevision(self, revno)
         return history[revno - 1]
 

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-08-16 19:15:19 +0000
+++ bzrlib/errors.py	2006-08-21 21:45:16 +0000
@@ -494,7 +494,6 @@
         self.format = format
 
 
-
 class StrictCommitFailed(Exception):
     """Commit refused because there are unknowns in the tree."""
 
@@ -505,8 +504,23 @@
     is_user_error = False
 
     def __init__(self, branch, revision):
-        self.branch = branch
-        self.revision = revision
+        BzrNewError.__init__(self, branch=branch, revision=revision)
+
+
+# Inheriting from NoSuchRevision to maintain API compatibility
+# technically, this could be a completely different exception
+class InvalidRevisionSpec(NoSuchRevision):
+    """Requested revision: %(spec)s does not exist in branch: %(branch)s%(extra)s"""
+
+    is_user_error = True
+
+    def __init__(self, spec, branch, extra=None):
+        NoSuchRevision.__init__(self, branch=branch, revision=spec)
+        self.spec = spec
+        if extra:
+            self.extra = '; ' + str(extra)
+        else:
+            self.extra = ''
 
 
 class HistoryMissing(BzrError):

=== modified file 'bzrlib/revisionspec.py'
--- bzrlib/revisionspec.py	2006-07-28 16:05:23 +0000
+++ bzrlib/revisionspec.py	2006-08-21 21:45:16 +0000
@@ -15,9 +15,13 @@
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 
+import bisect
 import datetime
 import re
-import bisect
+
+from bzrlib import (
+    errors,
+    )
 from bzrlib.errors import BzrError, NoSuchRevision, NoCommits
 
 _marker = []
@@ -182,6 +186,7 @@
 
 class RevisionSpec_int(RevisionSpec):
     """Spec is a number.  Special case."""
+
     def __init__(self, spec):
         self.spec = int(spec)
 
@@ -190,7 +195,10 @@
             revno = len(revs) + self.spec + 1
         else:
             revno = self.spec
-        rev_id = branch.get_rev_id(revno, revs)
+        try:
+            rev_id = branch.get_rev_id(revno, revs)
+        except NoSuchRevision:
+            raise errors.InvalidRevisionSpec(self.spec, branch)
         return RevisionInfo(branch, revno, rev_id)
 
 
@@ -199,21 +207,38 @@
 
     def _match_on(self, branch, revs):
         """Lookup a revision by revision number"""
-        if self.spec.find(':') == -1:
-            try:
-                return RevisionInfo(branch, int(self.spec))
-            except ValueError:
-                return RevisionInfo(branch, None)
-        else:
-            from branch import Branch
-            revname = self.spec[self.spec.find(':')+1:]
-            other_branch = Branch.open_containing(revname)[0]
-            try:
-                revno = int(self.spec[:self.spec.find(':')])
-            except ValueError:
-                return RevisionInfo(other_branch, None)
-            revid = other_branch.get_rev_id(revno)
-            return RevisionInfo(other_branch, revno)
+        loc = self.spec.find(':')
+        if loc == -1:
+            revno_spec = self.spec
+            branch_spec = None
+        else:
+            revno_spec = self.spec[:loc]
+            branch_spec = self.spec[loc+1:]
+
+        if revno_spec == '':
+            if branch_spec is None:
+                raise errors.InvalidRevisionSpec(branch, self.spec,
+                        'cannot have an empty revno and no branch')
+            revno = None
+        else:
+            try:
+                revno = int(revno_spec)
+            except ValueError, e:
+                raise errors.InvalidRevisionSpec(branch, self.spec, e)
+
+            if revno < 0:
+                revno = len(revs) + revno + 1
+
+        if branch_spec is not None:
+            from bzrlib.branch import Branch
+            branch = Branch.open(branch_spec)
+
+        try:
+            revid = branch.get_rev_id(revno)
+        except NoSuchRevision:
+            raise errors.InvalidRevisionSpec(branch, self.spec)
+
+        return RevisionInfo(branch, revno)
         
     def needs_branch(self):
         return self.spec.find(':') == -1

=== modified file 'bzrlib/tests/blackbox/test_annotate.py'
--- bzrlib/tests/blackbox/test_annotate.py	2006-07-28 16:05:23 +0000
+++ bzrlib/tests/blackbox/test_annotate.py	2006-08-21 21:45:16 +0000
@@ -89,7 +89,7 @@
         out, err = self.run_bzr_captured(['annotate', 'hello.txt', '-r 10'],
                                          retcode=3)
         self.assertEquals(out, '')
-        self.assertContainsRe(err, 'has no revision 10')
+        self.assertContainsRe(err, 'Requested revision: 10 does not exist')
 
     def test_annotate_cmd_two_revisions(self):
         out, err = self.run_bzr_captured(['annotate', 'hello.txt', '-r 1..2'],

=== modified file 'bzrlib/tests/test_revisionnamespaces.py'
--- bzrlib/tests/test_revisionnamespaces.py	2006-07-28 16:05:23 +0000
+++ bzrlib/tests/test_revisionnamespaces.py	2006-08-21 21:45:16 +0000
@@ -17,11 +17,13 @@
 import os
 import time
 
+from bzrlib import (
+    errors,
+    )
 from bzrlib.builtins import merge
 from bzrlib.branch import Branch
 from bzrlib.tests import TestCaseWithTransport
 from bzrlib.errors import NoCommonAncestor, NoCommits
-from bzrlib.errors import NoSuchRevision
 from bzrlib.revisionspec import RevisionSpec
 
 
@@ -74,13 +76,13 @@
                           (1, 'a@r-0-1'))
         self.assertEquals(RevisionSpec('revid:a@r-0-1').in_history(b),
                           (1, 'a@r-0-1'))
-        self.assertRaises(NoSuchRevision,
+        self.assertRaises(errors.NoSuchRevision,
                           RevisionSpec('revid:a@r-0-0').in_history, b)
         self.assertRaises(TypeError, RevisionSpec, object)
 
         self.assertEquals(RevisionSpec('date:today').in_history(b),
                           (2, 'a@r-0-2'))
-        self.assertRaises(NoSuchRevision,
+        self.assertRaises(errors.NoSuchRevision,
                           RevisionSpec('date:tomorrow').in_history, b)
         self.assertEquals(RevisionSpec('date:yesterday').in_history(b),
                           (1, 'a@r-0-1'))
@@ -133,3 +135,23 @@
         rev_info = spec.in_history(branch)
         self.assertEqual(rev_info, (None, 'A'))
 
+    def test_invalid_revno(self):
+        self.build_tree(['branch1/', 'branch1/file'])
+        wt = self.make_branch_and_tree('branch1')
+        wt.add('file')
+        wt.commit('first commit', rev_id='r1')
+        wt.commit('second commit', rev_id='r2')
+
+        # In the future -20 will probably just fall back to 0
+        # but for now, we want to make sure it raises the right error
+        self.assertRaises(errors.InvalidRevisionSpec,
+                          RevisionSpec('-20').in_history, wt.branch)
+        self.assertRaises(errors.InvalidRevisionSpec,
+                          RevisionSpec('10').in_history, wt.branch)
+
+        self.assertRaises(errors.InvalidRevisionSpec,
+                          RevisionSpec('revno:-20').in_history, wt.branch)
+        self.assertRaises(errors.InvalidRevisionSpec,
+                          RevisionSpec('revno:10').in_history, wt.branch)
+        self.assertRaises(errors.InvalidRevisionSpec,
+                          RevisionSpec('revno:a').in_history, wt.branch)

